import datetime
import json
import re
from asyncio import sleep, create_task
from io import BytesIO
from os import remove, path
from red_star.plugin_manager import BasePlugin
from discord import Embed, File, Forbidden, utils, Colour
from red_star.command_dispatcher import Command
from red_star.rs_errors import CommandSyntaxError, UserPermissionError, CustomCommandSyntaxError
from red_star.rs_utils import respond, find_user, decode_json, group_items
from .rs_lisp import lisp_eval, parse, reprint, standard_env, get_args
from dataclasses import dataclass, astuple
from subprocess import Popen, PIPE, TimeoutExpired
from sys import executable


@dataclass
class CCFileMetadata:
    owner: int
    size: int
    desc: str


# noinspection PyBroadException
class CustomCommands(BasePlugin):
    name = "custom_commands"
    version = "2.1"
    author = "GTG3000, medeor413"
    description = "A plugin that allows users to create custom commands using Red Star's " \
                  "custom RSLisp language dialect."
    default_config = {
        "default": {
            "cc_prefix": "!!",
            "cc_limit": 25
        },
        "rslisp_max_runtime": 5,
        "rslisp_minify": True,
        "cc_file_quota": 1024 * 1024,  # one megabyte
    }
    channel_categories = {"no_cc"}
    log_events = {"cc_event"}
    rpn_path = None

    async def activate(self):
        self.ccs = self.config_manager.get_plugin_config_file("ccs.json", json_save_args={'ensure_ascii': False})
        save_args = {'default': lambda o: astuple(o), 'ensure_ascii': False}
        load_args = {'object_pairs_hook': lambda obj: {k: CCFileMetadata(*v) for k, v in obj}}
        self.ccfdata = self.config_manager.get_plugin_config_file("cc_storage.json", json_save_args=save_args,
                                                                  json_load_args=load_args)
        self.ccfolder = self.client.storage_dir / "ccfiles"
        self.ccfolder.mkdir(parents=True, exist_ok=True)

        for rpn_exec in (directory / "rpn_executor.py" for directory in self.client.plugin_directories):
            if path.isfile(rpn_exec):
                self.rpn_path = rpn_exec
                break

        try:
            self.bans = self.ccs["bans"]
        except KeyError:
            self.bans = self.ccs["bans"] = {}

    # Event hooks

    async def on_message(self, msg):
        gid = str(msg.guild.id)
        self._initialize(gid)
        deco = self.plugin_config[gid]["cc_prefix"]
        if msg.author != self.client.user:
            cnt = msg.content
            if cnt.startswith(deco):
                if msg.author.id in self.bans[gid]["cc_use_ban"]:
                    try:
                        await msg.author.send(f"**WARNING: You are banned from usage of custom commands on this "
                                              f"server.**")
                    except Forbidden:
                        pass
                    return
                elif self.channel_manager.channel_in_category(msg.guild, "no_cc", msg.channel):
                    await self.plugin_manager.hook_event("on_log_event", msg.guild,
                                                         f"**WARNING: Attempted CC use in restricted channel"
                                                         f" {msg.channel.mention} by: {msg.author.display_name}**",
                                                         log_type="cc_event")
                    return

                cmd = cnt[len(deco):].split()[0].lower()

                if gid not in self.ccs:
                    self.ccs[gid] = {}
                if cmd in self.ccs[gid]:
                    if "restricted" not in self.ccs[gid][cmd]:
                        self.ccs[gid][cmd]["restricted"] = []
                    if self.ccs[gid][cmd]["restricted"]:
                        for t_cat in self.ccs[gid][cmd]["restricted"]:
                            if self.channel_manager.channel_in_category(msg.guild, t_cat, msg.channel):
                                break
                        else:
                            await self.plugin_manager.hook_event("on_log_event", msg.guild,
                                                                 f"**WARNING: Attempted CC use outside of it's "
                                                                 f"categories in {msg.channel.mention} by: "
                                                                 f"{msg.author}.**",
                                                                 log_type="cc_event")
                            return
                    await self.run_cc(cmd, msg)

    # Commands

    @Command("ReloadCCs",
             doc="Reloads custom commands from file.",
             category="custom_commands",
             bot_maintainers_only=True)
    async def _reloadccs(self, msg):
        self.ccs.reload()
        await respond(msg, "**AFFIRMATIVE. CCS reloaded.**")

    @Command("CreateCC", "NewCC",
             doc="Creates a custom command.\n"
                 "RSLisp Documentation: https://github.com/medeor413/Red_Star/wiki/Custom-Commands",
             syntax="(name) (content, in plain text or in an attached file)",
             category="custom_commands")
    async def _createcc(self, msg):
        gid = str(msg.guild.id)
        self._initialize(gid)
        if msg.author.id in self.bans[gid]["cc_create_ban"]:
            raise UserPermissionError("You are banned from creating custom commands.")
        if msg.attachments:
            fp = BytesIO()
            await msg.attachments[0].save(fp)
            args = msg.clean_content.split()
            name = args[1].lower() if len(args) > 1 else msg.attachments[0].filename.rsplit('.', 1)[0]
            content = fp.getvalue().decode()
        else:
            try:
                args = msg.clean_content.split(None, 2)[1:]
                name = args[0].lower()
            except IndexError:
                raise CommandSyntaxError("No name provided.")
            try:
                content = args[1]
            except IndexError:
                raise CommandSyntaxError("No content provided.")
        if gid not in self.ccs:
            self.ccs[gid] = {}
        if name in self.ccs[gid]:
            await respond(msg, f"**WARNING: Custom command {name} already exists.**")
        else:
            user_cc_count = len([True for cc in self.ccs[gid].values() if cc["author"] == msg.author.id])
            cc_limit = self.plugin_config[gid].get("cc_limit", 100)

            if msg.author.id not in self.config_manager.config.get("bot_maintainers", []) and not \
                    msg.author.permissions_in(msg.channel).manage_messages and user_cc_count >= cc_limit:
                raise UserPermissionError(f"Exceeded per-user custom command limit of {cc_limit}.")
            try:
                # check to see if there's something inside parenthesis floating in all the whitespace
                if not re.match(r"^\s*\(.*\)\s*$", content, re.DOTALL):
                    content = content.replace('"', '\\"')
                    content = f'"{content}"'
                parse(content)
            except Exception as err:
                await respond(msg, f"**WARNING: Custom command is invalid. Error: {err}**")
                return
            newcc = {
                "name": name,
                "content": reprint(parse(content)) if self.plugin_config['rslisp_minify'] else content,
                "author": msg.author.id,
                "date_created": datetime.datetime.now().strftime("%Y-%m-%d @ %H:%M:%S"),
                "last_edited": None,
                "locked": False,
                "restricted": [],
                "times_run": 0
            }
            self.ccs[gid][name] = newcc
            self.ccs.save()
            await respond(msg, f"**ANALYSIS: Custom command {name} created successfully.**")

    @Command("DumpCC",
             doc="Uploads the contents of the specified custom command as a text file.",
             syntax="(name)",
             category="custom_commands")
    async def _dumpcc(self, msg):
        gid = str(msg.guild.id)
        self._initialize(gid)
        if msg.author.id in self.bans[gid]["cc_create_ban"]:
            raise UserPermissionError("You are banned from editing custom commands.")
        try:
            name = msg.content.split(" ", 1)[1].lower()
        except IndexError:
            raise CommandSyntaxError("No name provided.")
        if name in self.ccs[gid]:
            async with msg.channel.typing():
                await respond(msg, "**AFFIRMATIVE. Completed file upload.**",
                              file=File(BytesIO(bytes(self.ccs[gid][name]["content"], encoding="utf-8")),
                                        filename=name + ".lisp"))
        else:
            raise CommandSyntaxError(f"No such custom command {name}.")

    @Command("EditCC",
             doc="Edits a custom command you created.",
             syntax="(name) (content, in plain text or in an attached file)",
             category="custom_commands")
    async def _editcc(self, msg):
        gid = str(msg.guild.id)
        self._initialize(gid)
        if msg.author.id in self.bans[gid]["cc_create_ban"]:
            raise UserPermissionError("You are banned from editing custom commands.")
        if msg.attachments:
            fp = BytesIO()
            await msg.attachments[0].save(fp)
            try:
                name = msg.clean_content.split(None, 2)[1].lower()
            except IndexError:
                name = msg.attachments[0].filename.rsplit('.', 1)[0]
            content = fp.getvalue().decode()
        else:
            try:
                _, name, content = msg.clean_content.split(" ", 2)
            except ValueError:
                raise CommandSyntaxError
        if gid not in self.ccs:
            self.ccs[gid] = {}
        if name in self.ccs[gid]:
            cc_data = self.ccs[gid][name]
            if cc_data["author"] == msg.author.id or msg.author.guild_permissions.manage_messages:
                try:
                    parse(content)
                except Exception as err:
                    await respond(msg, f"**WARNING: Custom command is invalid. Error: {err}**")
                    return
                cc_data["content"] = reprint(parse(content)) if self.plugin_config['rslisp_minify'] else content
                cc_data["last_edited"] = datetime.datetime.now().strftime("%Y-%m-%d @ %H:%M:%S")
                self.ccs[gid][name] = cc_data
                self.ccs.save()
                await respond(msg, f"**ANALYSIS: Custom command {name} edited successfully.**")
            else:
                raise UserPermissionError(f"You don't own custom command {name}.")
        else:
            await respond(msg, f"**WARNING: No such custom command {name}.**")

    @Command("DeleteCC", "DelCC", "RMCC",
             doc="Deletes a custom command.",
             syntax="(name)",
             category="custom_commands")
    async def _delcc(self, msg):
        gid = str(msg.guild.id)
        self._initialize(gid)
        if msg.author.id in self.bans[gid]["cc_create_ban"]:
            raise UserPermissionError("You are banned from deleting custom commands.")
        try:
            name = msg.clean_content.split(None, 1)[1].lower()
        except IndexError:
            raise CommandSyntaxError("No name provided.")
        if gid not in self.ccs:
            self.ccs[gid] = {}
        if name in self.ccs[gid]:
            if self.ccs[gid][name]["author"] == msg.author.id or msg.author.guild_permissions.manage_messages:
                del self.ccs[gid][name]
                self.ccs.save()
                await respond(msg, f"**ANALYSIS: Custom command {name} deleted successfully.**")
            else:
                raise UserPermissionError(f"You don't own custom command {name}.")
        else:
            await respond(msg, f"**WARNING: No such custom command {name}.**")

    @Command("CCInfo",
             doc="Displays information about a custom command.",
             syntax="(name)",
             category="custom_commands")
    async def _ccinfo(self, msg):
        try:
            name = msg.clean_content.split(None, 1)[1].lower()
        except IndexError:
            raise CommandSyntaxError("No name provided.")
        gid = str(msg.guild.id)
        if gid not in self.ccs:
            self.ccs[gid] = {}
        if name in self.ccs[gid]:
            cc_data = self.ccs[gid][name]
            last_edited = f"Last Edited: {cc_data['last_edited']}\n" if cc_data["last_edited"] else ""
            cc_locked = "Yes" if cc_data["locked"] else "No"
            author = utils.get(msg.guild.members, id=cc_data["author"])
            if author:
                author = str(author)
            else:
                author = "<Unknown user>"
            datastr = f"**ANALYSIS: Information for custom command {name}:**```\nName: {name}\nAuthor: {author}\n" \
                      f"Date Created: {cc_data['date_created']}\n{last_edited}Locked: {cc_locked}\n" \
                      f"Times Run: {cc_data['times_run']}```"
            await respond(msg, datastr)
        else:
            await respond(msg, f"**WARNING: No such custom command {name}.**")

    @Command("SearchCCs", "SearchCC", "ListCCs", "ListCC",
             doc="Searches CCs by name or author.",
             syntax="(name, author, or */all)",
             category="custom_commands")
    async def _searchccs(self, msg):
        search = msg.content.split(None, 1)[1].lower()
        if not search:
            raise CommandSyntaxError("No search provided.")
        user = find_user(msg.guild, search)
        gid = str(msg.guild.id)
        if gid not in self.ccs:
            self.ccs[gid] = {}
        ccs_list = list(self.ccs[gid].keys())
        if search in ("*", "all"):
            matched_ccs = ccs_list
        elif user:
            matched_ccs = filter(lambda x: self.ccs[gid][x]["author"] == user.id, ccs_list)
        else:
            matched_ccs = filter(lambda x: search in x.lower(), ccs_list)
        if matched_ccs:
            for split_msg in group_items(matched_ccs, "**ANALYSIS: The following custom commands match your "
                                                      "search:**", joiner=', '):
                await respond(msg, split_msg)

        else:
            await respond(msg, "**WARNING: No results found for your search.**")

    @Command("LockCC",
             doc="Toggles lock on a custom command, preventing it from being used.",
             syntax="(name)",
             category="custom_commands",
             perms={"manage_messages"})
    async def _lockcc(self, msg):
        try:
            name = msg.clean_content.split(None, 1)[1].lower()
        except IndexError:
            raise CommandSyntaxError("No name provided.")
        gid = str(msg.guild.id)
        if gid not in self.ccs:
            self.ccs[gid] = {}
        if name in self.ccs[gid]:
            self.ccs[gid][name]["locked"] = not self.ccs[gid][name]["locked"]
            lock_status = "locked" if self.ccs[gid][name]["locked"] else "unlocked"
            await respond(msg, f"**ANALYSIS: Custom command {name} has been {lock_status}.**")
        else:
            await respond(msg, f"**WARNING: No such custom command {name}.**")

    @Command("RestrictCC",
             doc="Restricts specified custom command to a specified category of channels, or removes said "
                 "restriction.",
             syntax="(name) (category)",
             category="custom_commands",
             perms={"manage_messages"})
    async def _restrictcc(self, msg):
        gid = str(msg.guild.id)
        if gid not in self.ccs:
            self.ccs[gid] = {}
        try:
            _, name, category = msg.content.split(None, 2)
        except ValueError:
            raise CommandSyntaxError("Two arguments required.")
        if name in self.ccs[gid]:
            whitelist = self.ccs[gid][name].setdefault("restricted", [])

            if self.channel_manager.get_category(msg.guild, category):
                if category not in whitelist:
                    whitelist.append(category)
                    await respond(msg, f"**AFFIRMATIVE. Custom command {name} restricted to category {category}.**")
                else:
                    whitelist.remove(category)
                    await respond(msg, f"**AFFIRMATIVE. Custom command {name} no longer restricted to category "
                                       f"{category}.**")
            else:
                raise CommandSyntaxError(f"No channel category by name of {category}.")
        else:
            raise CommandSyntaxError(f"No custom command by name of {name}.")

    @Command("CCMute", "MuteCC",
             doc="Toggles users ability to use custom commands.",
             syntax="(user)",
             category="custom_commands",
             perms={"manage_messages"})
    async def _mutecc(self, msg):
        gid = str(msg.guild.id)
        self._initialize(gid)
        user = find_user(msg.guild, msg.content.split(None, 1)[1])
        if not user:
            raise CommandSyntaxError("Not a user, or user not found.")
        if user.id in self.bans[gid]["cc_use_ban"]:
            self.bans[gid]["cc_use_ban"].remove(user.id)
            await respond(msg, f"**AFFIRMATIVE. User {user} was allowed the usage of custom commands.**")
        else:
            self.bans[gid]["cc_use_ban"].append(user.id)
            await respond(msg, f"**AFFIRMATIVE. User {user} was banned from using custom commands.**")

    @Command("CCBan", "BanCC",
             doc="Toggles users ability to create and alter custom commands.",
             syntax="(user)",
             category="custom_commands",
             perms={"manage_messages"})
    async def _bancc(self, msg):
        gid = str(msg.guild.id)
        self._initialize(gid)
        user = find_user(msg.guild, msg.content.split(None, 1)[1])
        if not user:
            raise CommandSyntaxError("Not a user, or user not found.")
        if user.id in self.bans[gid]["cc_create_ban"]:
            self.bans[gid]["cc_create_ban"].remove(user.id)
            await respond(msg, f"**AFFIRMATIVE. User {user} was allowed creation of custom commands.**")
        else:
            self.bans[gid]["cc_create_ban"].append(user.id)
            await respond(msg, f"**AFFIRMATIVE. User {user} was banned from creating custom commands.**")

    @Command("ListCCbans",
             doc="Lists users banned from using or creating CCs",
             syntax="(user)",
             category="custom_commands",
             perms={"manage_messages"})
    async def _listccban(self, msg):
        gid = str(msg.guild.id)
        self._initialize(gid)
        banned_users = {}
        for uid in self.bans[gid]["cc_create_ban"]:
            if uid in self.bans[gid]["cc_use_ban"]:
                banned_users[uid] = (True, True)
            else:
                banned_users[uid] = (True, False)
        for uid in self.bans[gid]["cc_use_ban"]:
            if uid not in banned_users:
                banned_users[uid] = (False, True)
        result_list = [f"{msg.guild.get_member(k).display_name:<32} | {str(v[0]):<5)} | "
                       f"{str(v[1]):<5}\n" for k, v in banned_users.items()]
        result_list.insert(0, f"{'Username'.ljust(32)} |  Ban  |  Mute")
        for split_msg in group_items(result_list, "**ANALYSIS: Currently banned members:**"):
            await respond(msg, split_msg)

    @Command("RPN",
             doc="Calculates an expression in extended reverse polish notation.\n"
                 "Binary operators: +, -, *, /, ^ (power), % (modulo), // (integer division), atan2, swap (swaps "
                 "two numbers in stack), log.\n"
                 "Unary operators: sin, cos, tan, ln, pop (remove number from stack), int, dup (duplicate number in "
                 "stack), drop, modf, round, rndint.\n"
                 "Constants: e, pi, tau, m2f (one meter in feet), m2i (one meter in inches), rnd.",
             run_anywhere=True)
    async def _rpncmd(self, msg):
        if self.rpn_path is None:
            return

        # sanitizing the input. Couldn't figure out a way to exploit Popen but *just in case*.
        num = re.compile(r"\d*\.\d+|\d+\.|0[xbo]\d+")
        ops = ("+", "-", "*", "/", "^", "%", "//", "log", "atan2", "swap", "min", "max", "sin", "cos", "tan", "ln",
               "pop", "int", "dup", "drop", "modf", "round", "rndint", "e", "pi", "tau", "m2f", "m2i", "rnd")

        args = [executable, str(self.rpn_path)] + [a for a in msg.content.lower().split()
                                                   if a in ops or a.isnumeric() or num.match(a)]

        # a convoluted way to run the RPN in such a way that plugging 3 3 3 3 ^ ^ ^ or something like that into the bot
        # doesn't make it lock up.
        # TODO: figure out if we can make multiprocessing work after all, this is kind of a hack.
        process = Popen(args, stdout=PIPE, stderr=PIPE, encoding="utf-8")
        try:
            output, err = process.communicate(timeout=self.plugin_config.get('rslisp_max_runtime', 5))
            process.wait()
            if err:
                raise CommandSyntaxError(output)
            result = output.split()
        except TimeoutExpired:
            process.kill()
            raise CommandSyntaxError("Command ran too long.")

        await respond(msg, f"**Result : [ {' | '.join([str(x) for x in result])} ]**")

    # Custom command machinery

    @Command("EvalCC",
             doc="Evaluates the given string through RSLisp cc parser.",
             syntax="(custom command)",
             category="custom_commands",
             perms={"manage_messages"})
    async def _evalcc(self, msg):
        program = msg.content.split(None, 1)[1]
        try:
            program = parse(program)
        except Exception as e:
            await respond(msg, f"**WARNING: Syntax error in custom command:** {e}")
        try:
            env = self._env(msg)
            result = lisp_eval(program, env)
        except Exception as e:
            await respond(msg, f"**WARNING: Runtime error in custom command:** {e}")
        else:
            if env['_rsoutput']:
                await respond(msg, str(env['_rsoutput']))
            elif result:
                await respond(msg, str(result))

    @Command("UploadCCData",
             doc="Uploads a cc-accessible data file in a json format.\n"
                 "File id must be one word with no dots/slashes that is up to 20 symbols.\n"
                 "Description is mandatory and needs to describe the contents of the file.\n"
                 "File itself must be valid JSON.",
             syntax="(file id) (description) +json file",
             category="custom_commands")
    async def _uploadccdata(self, msg):
        try:
            _, fid, desc = msg.clean_content.split(None, 2)
            fp = BytesIO()
            await msg.attachments[0].save(fp)

            # poor mans deflating. We're still storing JSON, but we want to store the least JSON possible.
            content = json.dumps(decode_json(fp.getvalue()), separators=(',', ':'), ensure_ascii=False)
        except json.decoder.JSONDecodeError as e:
            raise CommandSyntaxError(e)
        except (ValueError, IndexError):
            raise CommandSyntaxError

        fid = fid.lower()[:20]

        # since the fid string will just be slapped onto the end of a path, it's a good idea to make sure users
        # can't just point at whatever they see fit and go "that's my file"
        if not fid.isidentifier():
            raise CommandSyntaxError("Illegal characters in file ID.")

        total_size = sum([x.size for x in self.ccfdata.values() if x.owner == msg.author.id])

        if fid in self.ccfdata:
            if self.ccfdata[fid].owner != msg.author.id:
                raise UserPermissionError
            total_size -= self.ccfdata[fid].size

        size = len(content)
        if total_size + size > self.plugin_config['cc_file_quota'] \
                and not self.config_manager.is_maintainer(msg.author):
            raise UserPermissionError("File exceeds size quota. Remaining quota: "
                                      f"{self.plugin_config['cc_file_quota']-total_size} bytes.")

        self.ccfdata[fid] = CCFileMetadata(msg.author.id, size, desc.replace('\r', ''))

        total_size = self.plugin_config['cc_file_quota'] - total_size - size

        with (self.ccfolder / (fid + '.json')).open('w', encoding='utf8') as fp:
            fp.write(content)

        await respond(msg, f"**AFFIRMATIVE. CC data file {fid} now available.\nRemaining quota: {total_size} bytes.**")

    @Command("ListCCData",
             doc="Prints a list of all available data files, including first line of description (shortened to 50 "
                 "characters), owner and size.",
             category="custom_commands")
    async def _listccdata(self, msg):
        # since the descriptions may be multiline, it's nice to remove any possible newlines and what comes after.
        def desc(v: CCFileMetadata):
            return v.desc.split('\n')[0][:50]

        items = (f"{k:20} : {desc(v):50} : {str(self.client.get_user(v.owner))} : {v.size} bytes"
                 for k, v in self.ccfdata.items())

        for split in group_items(items, "**ANALYSIS: Following data files available:**"):
            await respond(msg, split)

    @Command("DeleteCCData",
             doc="Removes a data file.",
             syntax="(file id)",
             category="custom_commands")
    async def _delccdata(self, msg):
        try:
            fid = msg.clean_content.split(None, 1)[1].lower()[:20]

            if fid in self.ccfdata:
                if self.ccfdata[fid].owner != msg.author.id and not self.config_manager.is_maintainer(msg.author):
                    raise UserPermissionError("File belongs to another user.")
                del self.ccfdata[fid]
                remove(self.ccfolder / (fid + '.json'))
                await respond(msg, f"**AFFIRMATIVE. File {fid} removed.**")
            else:
                raise CommandSyntaxError(f"No file {fid} found.")
        except IndexError:
            raise CommandSyntaxError
        except OSError:
            raise CommandSyntaxError("Error while deleting file. Please contact the bot maintainer.")

    @Command("CCData",
             doc="Prints out the information associated with a data file, including person uploading, "
                 "size and description.",
             syntax="(file id)",
             category="custom_commands")
    async def _ccdata(self, msg):
        try:
            fid = msg.clean_content.split()[1].lower()[:20]
            if fid in self.ccfdata:
                f = self.ccfdata[fid]
                await respond(msg, f"`{fid} uploaded by {self.client.get_user(f.owner)} ({f.size}b)`\n\n{f.desc}")
            else:
                raise CommandSyntaxError(f"No file {fid} found.")
        except IndexError:
            raise CommandSyntaxError

    @Command("EditCCData",
             doc="Changes the description associated to one given.",
             syntax="(file id) (description)",
             category="custom_commands")
    async def _editccdata(self, msg):
        try:
            _, fid, desc = msg.clean_content.split(None, 2)
            if fid in self.ccfdata:
                if self.ccfdata[fid].owner != msg.author.id and not self.config_manager.is_maintainer(msg.author):
                    raise UserPermissionError
                self.ccfdata[fid].desc = desc.replace('\r', '')
                await respond(msg, "**AFFIRMATIVE. Description updated.**")
            else:
                raise CommandSyntaxError(f"No file {fid} found.")
        except ValueError:
            raise CommandSyntaxError

    @Command("DumpCCData",
             doc="Uploads the specified data file",
             syntax="(file id)",
             category="custom_commands")
    async def _dumpccdata(self, msg):
        try:
            _, fid = msg.clean_content.split(None, 1)
            fid = fid.lower()[:20]
            if fid in self.ccfdata:
                async with msg.channel.typing():
                    await respond(msg, "**AFFIRMATIVE.**",
                                  file=File((self.ccfolder/(fid+'.json')).resolve().as_posix(),
                                            filename=fid + ".json"))
            else:
                raise CommandSyntaxError(f"No file {fid} found.")
        except ValueError:
            raise CommandSyntaxError

    @staticmethod
    async def _rm_msg(msg):
        await sleep(1)
        await msg.delete()

    def _initialize(self, gid):
        if gid not in self.plugin_config:
            self.plugin_config[gid] = self.default_config["default"].copy()
            self.config_manager.save_config()
        if gid not in self.bans:
            self.bans[gid] = {
                "cc_create_ban": [],
                "cc_use_ban": []
            }

    async def run_cc(self, cmd, msg):
        gid = str(msg.guild.id)
        if self.ccs[gid][cmd]["locked"] and not msg.author.guild_permissions.manage_messages:
            await respond(msg, f"**WARNING: Custom command {cmd} is locked.**")
        else:
            env = self._env(msg)

            cc_data = self.ccs[gid][cmd]["content"]
            try:
                res = lisp_eval(parse(cc_data), env)
            except CustomCommandSyntaxError as e:
                err = e if e else "Syntax error."
                await respond(msg, f"**WARNING: Author made syntax error: {err}**")
            except CommandSyntaxError as e:
                err = e if e else "Syntax error."
                await respond(msg, f"**WARNING: {err}**")
            except Exception as e:
                err = e if e else "Syntax error."
                self.logger.exception("Exception occurred in custom command: ", exc_info=True)
                await respond(msg, f"**WARNING: An error occurred while running the custom command: {err}**")
            else:
                if env['_rsoutput']:
                    await respond(msg, env['_rsoutput'])
                elif res:
                    await respond(msg, str(res))
                self.ccs[gid][cmd]["times_run"] += 1
                self.ccs.save()

    #  tag functions that *require* the discord machinery

    def _env(self, msg):
        gid = str(msg.guild.id)
        cmd = msg.content[len(self.plugin_config[gid]["cc_prefix"]):].split()[0].lower()
        env = standard_env(max_runtime=self.plugin_config.get('rslisp_max_runtime', 0))

        env['username'] = msg.author.name
        env['usernick'] = msg.author.display_name
        env['usermention'] = msg.author.mention
        try:
            author = utils.get(msg.guild.members, id=self.ccs[gid][cmd]['author'])
            env['authorname'] = author.name
            env['authornick'] = author.display_name
        except (AttributeError, KeyError):
            env['authorname'] = env['authornick'] = '<Unknown user>'
        args = msg.clean_content.split(" ", 1)
        env['argstring'] = args[1] if len(args) > 1 else ''
        env['args'] = args[1].split(" ") if len(args) > 1 else []

        env['hasrole'] = lambda *x: self._hasrole(msg, *x)
        env['delcall'] = lambda: self._delcall(msg)
        env['embed'] = lambda *x: self._embed(msg, *get_args(x))
        env['file'] = lambda x: self._file(x)

        return env

    def _delcall(self, msg):
        create_task(self._rm_msg(msg))

    @staticmethod
    def _hasrole(msg, *args):
        _args = map(str.lower, args)
        return any([x.name.lower() in _args for x in msg.author.roles])

    @staticmethod
    def _embed(msg, _, kwargs):
        embed = Embed(type="rich", colour=16711680)
        can_post = False
        for name, value in kwargs.items():
            can_post = True
            if name.lower() == "!title":
                embed.title = value
            elif name.lower() in ["!color", "!colour"]:
                try:
                    embed.colour = value if isinstance(value, int) else Colour(int(value, 16))
                except ValueError:
                    pass
            elif name.lower() == "!url":
                embed.url = value
            elif name.lower() == "!thumbnail":
                embed.set_thumbnail(url=value)
            elif name.lower() == "!image":
                embed.set_image(url=value)
            elif name.lower() in ["!desc", "!description"]:
                embed.description = value
            elif name.lower() == "!footer":
                embed.set_footer(text=value)
            else:
                if type(value) == list:
                    content = value[0]
                    is_inline = value[1]
                else:
                    content = value
                    is_inline = False
                embed.add_field(name=name, value=content, inline=is_inline)
        if can_post:
            create_task(respond(msg, None, embed=embed))

    def _file(self, filename):
        if not filename.isidentifier():
            raise CustomCommandSyntaxError(f"file error: illegal filename {filename}")
        try:
            with (self.ccfolder / (filename + '.json')).open() as fp:
                return json.load(fp)
        except FileNotFoundError:
            raise CustomCommandSyntaxError(f"file error: no file named {filename}")
        except json.decoder.JSONDecodeError:
            raise CustomCommandSyntaxError(f"json error: file {filename} is invalid json")



import shlex
from string import capwords
from discord import InvalidArgument, HTTPException, Colour
from red_star.plugin_manager import BasePlugin
from red_star.rs_errors import CommandSyntaxError
from red_star.rs_utils import respond, is_positive, find_role, group_items, RSArgumentParser
from red_star.command_dispatcher import Command


class RoleCommands(BasePlugin):
    name = "role_commands"
    version = "1.1"
    author = "GTG3000"
    description = "A plugin for manipulating server roles via commands."

    @Command("EditRole",
             perms={"manage_roles"},
             category="roles",
             syntax="(role) [-n/--name string][-c/--colour FFFFFF][-h/--hoist bool][-m/--mentionable bool]"
                    "[-p/--position integer].\nANALYSIS: Strings can be encapsulated in \"...\" to allow spaces",
             doc="Edits the specified role name, colour, hoist (show separately from others) "
                 "and mentionable properties.\nOptions must be specified as \"--option value\" or \"-o value\".\n"
                 "Colour can be reset by setting it to 0.")
    async def _edit_role(self, msg):
        """
        a command for editing a role.
        !editrole (role name) [name=name][colour=colour][hoist=hoist][mentionable=mentionable]
        name is a string
        colour is a colour object (value converted from hexadecimal string)
        hoist and mentionable are boolean
        """
        try:
            args = shlex.split(msg.content)
        except ValueError as e:
            self.logger.warning(f"Unable to split {msg.content}. {e}")
            raise CommandSyntaxError(e)

        parser = RSArgumentParser()
        parser.add_argument("command")                      # Well it's gonna be there.
        parser.add_argument("role")                         # The role name/ID
        parser.add_argument("-n", "--name")                 # New role name
        parser.add_argument("-c", "--colour", "--color")    # New role colour
        parser.add_argument("-h", "--hoist")                # To separate the role or not
        parser.add_argument("-m", "--mentionable")          # To allow role being mentioned
        parser.add_argument("-p", "--position", type=int)   # Changing position (DON'T ACTUALLY USE IT)
        if len(args) > 1:
            args = parser.parse_args(args)

            role = find_role(msg.guild, args['role'])
            if role:
                try:
                    arg_dict = {
                        "name": args['name'] if args['name'] else None,
                        "colour": Colour(int(args['colour'], 16)) if args['colour'] else None,
                        "hoist": is_positive(args['hoist']) if args['hoist'] else None,
                        "mentionable": is_positive(args['mentionable']) if args['mentionable'] else None,
                        "position": max(0, args['position']) if args['position'] else None
                    }
                    arg_dict = {k: v for k, v in arg_dict.items() if v is not None}
                except ValueError:
                    raise CommandSyntaxError("Colour must be in web-colour hexadecimal format.")
                await role.edit(**arg_dict)
                result_string = "\n".join([f"{k}: {v}" for k, v in arg_dict.items()])
                await respond(msg, f"**AFFIRMATIVE. Role {role.name} modified with parameters :**\n"
                                   f"```{result_string}```")
            else:
                raise CommandSyntaxError(f"No role \"{args['role']}\" found.")
        else:
            raise CommandSyntaxError

    @Command("CreateRole",
             perms={"manage_roles"},
             category="roles",
             syntax="(role name) (base role) [-n/--name string][-c/--colour FFFFFF][-h/--hoist bool]"
                    "[-m/--mentionable bool][-p/--position integer].\n"
                    "ANALYSIS: Strings can be encapsulated in \"...\" to allow spaces",
             doc="Creates a role based on an existing role (for position and permissions), "
             "with parameters similar to editrole")
    async def _createrole(self, msg):
        """
        a command for creating a role
        takes names for new role and a role that will be copied for position/permissions
        """
        try:
            args = shlex.split(msg.content)
        except ValueError as e:
            self.logger.warning(f"Unable to split {msg.content}. {e}")
            raise CommandSyntaxError(e)

        parser = RSArgumentParser()
        parser.add_argument("command")                      # Well it's gonna be there.
        parser.add_argument("role")                         # Name of the new role
        parser.add_argument("template")                     # Permission donor name/ID
        parser.add_argument("-n", "--name")                 # New role name
        parser.add_argument("-c", "--colour", "--color")    # New role colour
        parser.add_argument("-h", "--hoist")                # To separate the role or not
        parser.add_argument("-m", "--mentionable")          # To allow role being mentioned
        parser.add_argument("-p", "--position", type=int)   # Changing position (DON'T ACTUALLY USE IT)

        if len(args) > 2:
            parsed_args = parser.parse_args(args)
            role = find_role(msg.guild, parsed_args['template'])
            if role:
                try:
                    arg_dict = {
                        "name": parsed_args['name'] if parsed_args['name'] else args[1],
                        "permissions": role.permissions,
                        "colour": Colour(int(parsed_args['colour'], 16)) if parsed_args['colour'] else role.colour,
                        "hoist": is_positive(parsed_args['hoist']) if parsed_args['hoist'] else role.hoist,
                        "mentionable": is_positive(parsed_args['mentionable'])
                        if parsed_args['mentionable'] else role.mentionable
                    }
                except ValueError:
                    raise CommandSyntaxError("Colour must be in web-colour hexadecimal format.")

                rolepos = max(0, parsed_args['position']) if parsed_args['position'] else role.position

                t_role = await msg.guild.create_role(**arg_dict)

                try:
                    # since I can't create a role with a preset position :T
                    await t_role.edit(position=rolepos)
                except (InvalidArgument, HTTPException):
                    # oh hey, why are we copying this role again?
                    name = args[1].capitalize()
                    await t_role.delete()
                    raise CommandSyntaxError(f"Failed to move role {name} to position {rolepos}.")
                t_string = ""
                for k, v in arg_dict.items():
                    if k != "permissions":
                        t_string = f"{t_string}{k}: {v!s}\n"
                    else:
                        t_string += k + ": " + ", ".join({x.upper() for x, y in v if y}) + "\n"
                name = args[1].capitalize()
                await respond(msg, f"**AFFIRMATIVE. Created role {name} with parameters :**\n ```{t_string}```")
            else:
                await respond(msg, f"**NEGATIVE. ANALYSIS: no base role {args[2].capitalize()} found.**")
        else:
            raise CommandSyntaxError

    @Command("DeleteRole",
             perms={"manage_roles"},
             category="roles",
             syntax="(role) [position].\nANALYSIS: Strings can be encapsulated in \"...\" to allow spaces",
             doc="Deletes first encounter of the role with the given name and optionally position.")
    async def _deleterole(self, msg):
        try:
            args = shlex.split(msg.content)
        except ValueError as e:
            self.logger.warning(f"Unable to split {msg.content}. {e}")
            raise CommandSyntaxError(e)
        if len(args) > 1:
            name = args[1].capitalize()
            pos = -1
            if len(args) > 2:
                try:
                    pos = int(args[2])
                except ValueError:
                    raise SyntaxWarning("Position should be integer.")
            role = find_role(msg.guild, args[1])
            if role and (((pos >= 0) and (role.position == pos)) or pos < 0):
                # delete if name matches, and if pos is not -1 - if position matches
                t_position = role.position
                await role.delete()
                await respond(msg, f"**AFFIRMATIVE. Deleted role: {name} in position: {str(t_position)}.**")
            else:
                await respond(msg, f"**NEGATIVE. ANALYSIS: no role {name} found.**")
        else:
            raise CommandSyntaxError("Expected role name.")

    @Command("RoleInfo", "InfoRole",
             category="roles",
             syntax="(role).\nANALYSIS: Strings can be encapsulated in \"...\" to allow spaces",
             doc="Returns all the info about the given role.")
    async def _inforole(self, msg):
        """
        provides an infodump of a role, including permissions and position
        """
        try:
            args = shlex.split(msg.content)
        except ValueError as e:
            self.logger.warning("Unable to split {data.content}. {e}")
            raise CommandSyntaxError(e)

        if len(args) > 1:
            name = capwords(args[1])
            role = find_role(msg.guild, args[1])
            if role:
                info_dict = {
                    "name": role.name,
                    "permissions": role.permissions,
                    "colour": role.colour,
                    "hoist": role.hoist,
                    "mentionable": role.mentionable,
                    "position": role.position,
                    "created_at": role.created_at,
                    "id": role.id
                }
                result_string = ""
                for k, v in info_dict.items():
                    if k != "permissions":
                        result_string = f"{result_string}{k}: {v!s}\n"
                    else:
                        result_string += k + ": " + ", ".join({x.upper() for x, y in v if y}) + "\n"
                await respond(msg, f"**ANALYSIS: role {role.name} has parameters :**\n ```{result_string}```")
            else:
                await respond(msg, f"**NEGATIVE. ANALYSIS: no role {name} found.**")
        else:
            raise CommandSyntaxError

    @Command("ListRoles", "ListRole",
             category="roles",
             perms={"manage_roles"},
             doc="Lists all roles.")
    async def _listroles(self, msg):
        """
        lists all roles along with position and color
        """
        role_list = (f"[{role.position:03d} | {role.colour} | {role.name[:40].ljust(40,'·')}]"
                     for role in msg.guild.roles[::-1])
        for split_msg in group_items(role_list, message="**AFFIRMATIVE. Listing roles :**"):
            await respond(msg, split_msg)



import sys
import setuptools
from red_star.rs_version import version


def get_requirements():
    with open("requirements.txt", encoding="utf-8") as fd:
        return fd.read()


def long_description():
    with open("README.md", encoding="utf-8") as fd:
        return fd.read()


setuptools.setup(
    name='red_star',
    version=version,
    description='Red Star - A general-purpose Discord bot with bonus shouting.',
    url='https://github.com/medeor413/Red_Star',
    author="Medeor",
    author_email="me@medeor.me",
    long_description=long_description(),
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    package_data={'red_star': ['_default_files/*.json']},
    license='MIT',
    platforms='any',
    python_requires="~=3.7",
    install_requires=get_requirements(),
    entry_points={
        'console_scripts': ['red_star = red_star.__main__:main']
    },
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Topic :: Communications :: Chat"
    ]
)



from datetime import datetime, timedelta
from discord import AuditLogAction, Forbidden
from red_star.plugin_manager import BasePlugin
from red_star.rs_errors import ChannelNotFoundError, CommandSyntaxError
from red_star.rs_utils import split_message, respond, close_markdown
from red_star.command_dispatcher import Command


class DiscordLogger(BasePlugin):
    name = "logger"
    version = "1.5"
    author = "medeor413, GTG3000"
    description = "A plugin that logs certain events and prints them to a defined log channel " \
                  "in an easily-readable manner."
    default_config = {
        "default": {
            "log_event_blacklist": [
            ]
        }
    }
    channel_types = {"logs"}
    log_events = {"message_delete", "message_edit", "member_update", "pin_update", "member_ban", "member_unban",
                  "member_join", "member_leave", "role_update"}

    async def activate(self):
        self.log_items = {}

    async def on_all_plugins_loaded(self):
        for plg in self.plugins.values():
            if plg is self:
                continue
            plg_log_events = getattr(plg, "log_events", None)
            if plg_log_events:
                self.log_events |= plg_log_events
                self.logger.debug(f"Registered log events {', '.join(plg_log_events)} from {plg.name}.")

    async def on_global_tick(self, *_):
        for guild in self.client.guilds:
            gid = str(guild.id)
            try:
                log_channel = self.channel_manager.get_channel(guild, "logs")
            except ChannelNotFoundError:
                continue
            if gid in self.log_items and self.log_items[gid]:
                logs = "\n".join(self.log_items[gid])
                if not logs:
                    continue
                for msg in split_message(logs, splitter="\n"):
                    if msg and not msg.isspace():
                        await log_channel.send(msg)
                self.log_items[gid].clear()

    async def on_message_delete(self, msg):
        blacklist = self.plugin_config.setdefault(str(msg.guild.id),
                                                  self.plugin_config["default"])["log_event_blacklist"]
        if "message_delete" not in blacklist and msg.author != self.client.user:
            contents, _ = close_markdown(msg.clean_content if msg.clean_content else msg.system_content)
            msgtime = msg.created_at.strftime("%Y-%m-%d @ %H:%M:%S")
            attaches = ""
            if msg.attachments:
                links = ", ".join([x.proxy_url or x.url for x in msg.attachments])
                attaches = f"\n**Attachments:** `{links}`"
            self.emit_log(f"**ANALYSIS: User {msg.author}'s message at `{msgtime}` in {msg.channel.mention}"
                          f" was deleted. ANALYSIS: Contents:**\n{contents}{attaches}", msg.guild)

            self.logger.info(f"{msg.author}'s message at {msgtime} in {msg.channel} of {msg.guild} was deleted:\n"
                             f"Contents:\n{contents}{attaches.replace('**','')}")

    async def on_message_edit(self, before, after):
        blacklist = self.plugin_config.setdefault(str(after.guild.id),
                                                  self.plugin_config["default"])["log_event_blacklist"]
        if "message_edit" not in blacklist and after.author != self.client.user:
            old_contents, _ = close_markdown(before.clean_content)
            contents, _ = close_markdown(after.clean_content)
            if old_contents == contents:
                return
            msgtime = after.created_at.strftime("%Y-%m-%d @ %H:%M:%S")
            self.emit_log(f"**ANALYSIS: User {after.author} edited their message at `{msgtime}` in "
                          f"{after.channel.mention}. ANALYSIS:**\n**Old contents:** {old_contents}\n"
                          f"**New contents:** {contents}", after.guild)

            self.logger.info(f"User {after.author} edited their message "
                             f"at {msgtime} in {after.channel} of {after.guild}.\n"
                             f"Old contents:\n{old_contents}\nNew contents:\n{contents}")

    async def on_member_update(self, before, after):
        blacklist = self.plugin_config.setdefault(str(after.guild.id),
                                                  self.plugin_config["default"])["log_event_blacklist"]
        if "member_update" not in blacklist:
            diff_str = log_str = ""
            if before.name != after.name or before.discriminator != after.discriminator:
                diff_str = f"`Old username: `{before}\n`New username: `{after}\n"
                log_str = f"Old username: {before}\nNew username: {after}\n"
            if before.avatar != after.avatar:
                diff_str = f"{diff_str}`New avatar: `{after.avatar_url}\n"
                log_str = f"{log_str}New avatar: {after.avatar_url}\n"
            if before.nick != after.nick:
                diff_str = f"{diff_str}`Old nick: `{before.nick}\n`New nick: `{after.nick}\n"
                log_str = f"{log_str}Old nick: {before.nick}\nNew nick: {after.nick}\n"
            if before.roles != after.roles:
                old_roles = ", ".join([str(x) for x in before.roles])
                new_roles = ", ".join([str(x) for x in after.roles])
                diff_str = f"{diff_str}**Old roles:**```[ {old_roles} ]```\n**New roles:**```[ {new_roles} ]```"
                log_str = f"{log_str}Old roles: [ {old_roles} ]\nNew roles: [ {new_roles} ]"
            if not diff_str:
                return
            self.emit_log(f"**ANALYSIS: User {after} was modified:**\n{diff_str}", after.guild)
            self.logger.info(f"User {after} was modified:\n{log_str}")

    async def on_guild_channel_pins_update(self, channel, last_pin):
        blacklist = self.plugin_config.setdefault(str(channel.guild.id),
                                                  self.plugin_config["default"])["log_event_blacklist"]
        if "pin_update" not in blacklist:
            cnt = None
            try:
                new_pin = (datetime.utcnow() - last_pin < timedelta(seconds=5))
            except TypeError:  # last_pin can be None if the last pin in a channel was unpinned
                new_pin = False
            if new_pin:  # Get the pinned message if it's a new pin; can't get the unpinned messages sadly
                msg = (await channel.pins())[0]
                cnt = msg.author, msg.clean_content
            self.emit_log(f"**ANALYSIS: A message was {'' if new_pin else 'un'}pinned in {channel.mention}.**\n"
                          f"{f'**Message: {cnt[0]}:** {cnt[1]}' if new_pin else ''}", channel.guild)
            self.logger.info(f"A message was {'' if new_pin else 'un'}pinned in {channel} of {channel.guild}\n"
                             f"{f'Message: {cnt[0]}: {cnt[1]}' if new_pin else ''}")

    async def on_member_ban(self, guild, member):
        blacklist = self.plugin_config.setdefault(str(guild.id), self.plugin_config["default"])["log_event_blacklist"]
        if "member_ban" not in blacklist:
            self.emit_log(f"**ANALYSIS: User {member} was banned.**", guild)
            self.logger.info(f"User {member} was benned in {guild}.")

    async def on_member_unban(self, guild, member):
        blacklist = self.plugin_config.setdefault(str(guild.id), self.plugin_config["default"])["log_event_blacklist"]
        if "member_unban" not in blacklist:
            self.emit_log(f"**ANALYSIS: Ban was lifted from user {member}.**", guild)
            self.logger.info(f"Ban was lifted from user {member} in {guild}")

    async def on_member_join(self, member):
        blacklist = self.plugin_config.setdefault(str(member.guild.id),
                                                  self.plugin_config["default"])["log_event_blacklist"]
        if "member_join" not in blacklist:
            self.emit_log(f"**ANALYSIS: User {member} has joined the server. User id: `{member.id}`**", member.guild)
            self.logger.info(f"User {member} has joined {member.guild}. User id: {member.id}.")

    async def on_member_remove(self, member):
        blacklist = self.plugin_config.setdefault(str(member.guild.id),
                                                  self.plugin_config["default"])["log_event_blacklist"]
        if "member_leave" not in blacklist:
            try:
                # find audit log entries for kicking of member with our ID, created in last five seconds.
                # Hopefully five seconds is enough
                latest_logs = member.guild.audit_logs(action=AuditLogAction.kick, limit=1)
                kick_event = await latest_logs.get(target__id=member.id)
            except Forbidden:
                kick_event = None
            if kick_event:
                kicker = kick_event.user
                reason_str = f"Reason: {kick_event.reason}; " if kick_event.reason else ""
                self.emit_log(f"**ANALYSIS: User {member} was kicked from the server by {kicker}. "
                              f"{reason_str}User id: `{member.id}`**", member.guild)
                self.logger.info(f"User {member} was kicked from {member.guild} by {kicker}. "
                                 f"{reason_str}User ud: {member.id}")
            else:
                self.emit_log(f"**ANALYSIS: User {member} has left the server. User id: `{member.id}`**", member.guild)
                self.logger.info(f"User {member} has left {member.guild}. User id: {member.id}.")

    async def on_guild_role_update(self, before, after):
        blacklist = self.plugin_config.setdefault(str(after.guild.id),
                                                  self.plugin_config["default"])["log_event_blacklist"]
        if "role_update" not in blacklist:
            diff = []
            try:
                audit_event = await after.guild.audit_logs(action=AuditLogAction.role_update, limit=1).get()
            except Forbidden:
                audit_event = None

            if before.name == after.name \
                    and before.colour == after.colour \
                    and before.hoist == after.hoist \
                    and before.mentionable == after.mentionable  \
                    and before.permissions == after.permissions \
                    and before.position == after.position:
                if audit_event is not None and audit_event.target == after:
                    before_dict = audit_event.changes.before.__dict__
                    before.name = before_dict.get("name", after.name)
                    before.colour = before_dict.get("colour", after.colour)
                    before.hoist = before_dict.get("hoist", after.hoist)
                    before.mentionable = before_dict.get("mentionable", after.mentionable)
                    before.permissions = before_dict.get("permissions", after.permissions)
                    before.position = before_dict.get("position", after.position)
                else:
                    return

            if before.name != after.name:
                diff.append(f"Name changed from {before.name} to {after.name}")
            if before.position != after.position:
                diff.append(f"Position changed from {before.position} to {after.position}")
            if before.colour != after.colour:
                diff.append(f"Colour changed from {before.colour} to {after.colour}")
            if before.hoist != after.hoist:
                diff.append("Is now displayed separately." if after.hoist else "Is no longer displayed separately.")
            if before.mentionable != after.mentionable:
                diff.append("Can now be mentioned." if after.mentionable else "Can no longer be mentioned.")
            if before.permissions != after.permissions:
                # comparing both sets of permissions, PITA
                before_perms = [x for x, y in before.permissions if y]
                after_perms = [x for x, y in after.permissions if y]
                perm_diff = "Added permissions: " + ", ".join(x.upper() for x in after_perms if x not in before_perms)
                perm_diff += "\nRemoved permissions: " + \
                             ", ".join(x.upper() for x in before_perms if x not in after_perms)
                diff.append(perm_diff)

            if not diff or (len(diff) == 1 and before.position != after.position):
                return

            diff = '\n'.join(diff)

            self.emit_log(f"**ANALYSIS: Role {before.name} was changed by "
                          f"{audit_event.user if audit_event else 'someone'}:**\n"
                          f"```\n{diff}```\n", after.guild)
            self.logger.info(f"Role {before.name} was changed by "
                             f"{audit_event.user if audit_event else 'someone'}:\n"
                             f"{diff}")

    async def on_log_event(self, guild, string, *, log_type="log_event"):
        blacklist = self.plugin_config.setdefault(str(guild.id), self.plugin_config["default"])["log_event_blacklist"]
        if log_type not in blacklist:
            self.emit_log(string, guild)
            self.logger.info(string)

    def emit_log(self, log_str, guild):
        guild_log_queue = self.log_items.setdefault(str(guild.id), [])
        guild_log_queue.append(log_str)

    @Command("LogEvent",
             doc="Adds or removes the events to be logged.",
             syntax="[add|remove type]",
             category="bot_management",
             perms={"manage_guild"})
    async def _logevent(self, msg):
        cfg = self.plugin_config.setdefault(str(msg.guild.id), self.plugin_config["default"])["log_event_blacklist"]
        try:
            action, event_type = msg.clean_content.lower().split(" ", 2)[1:]
            if event_type not in self.log_events:
                await respond(f"**Log event {event_type} does not exist.\n"
                              f"Valid events:** `{', '.join(self.log_events)}`")
                return
        except ValueError:
            if len(msg.clean_content.split(" ")) == 1:
                enabled = ", ".join(self.log_events - set(cfg)) or "None"
                await respond(msg, f"**ANALYSIS: Enabled log events:** `{enabled}`\n"
                                   f"**Disabled log events:** {', '.join(cfg) or 'None'}")
                return
            else:
                raise CommandSyntaxError("Invalid number of arguments.")
        if action == "remove":
            if event_type not in cfg:
                cfg.append(event_type)
                self.config_manager.save_config()
                await respond(msg, f"**ANALYSIS: No longer logging events of type {event_type}.**")
            else:
                await respond(msg, f"**ANALYSIS: Event type {event_type} is already disabled.**")
        elif action == "add":
            if event_type in cfg:
                cfg.remove(event_type)
                self.config_manager.save_config()
                await respond(msg, f"**ANALYSIS: Now logging events of type {event_type}.**")
            else:
                await respond(msg, f"**ANALYSIS: Event type {event_type} is already logged.**")
        else:
            raise CommandSyntaxError(f"Action {action} is not a valid action.")



import logging
from asyncio import create_task, sleep
from datetime import datetime
from discord import AutoShardedClient, DMChannel, Intents
from discord.object import Object
from discord.utils import oauth_url
from pathlib import Path
from sys import exc_info
from red_star.channel_manager import ChannelManager
from red_star.command_dispatcher import CommandDispatcher
from red_star.config_manager import ConfigManager
from red_star.plugin_manager import PluginManager


class RedStar(AutoShardedClient):
    def __init__(self, storage_dir, argv):
        self.logger = logging.getLogger("red_star")
        dpy_logger = logging.getLogger("discord")
        if argv.verbose > 0:
            self.logger.setLevel(logging.DEBUG)
            dpy_logger.setLevel(logging.DEBUG if argv.verbose >= 2 else logging.INFO)
        else:
            self.logger.setLevel(logging.INFO)
            dpy_logger.setLevel(logging.INFO)
        self.logger.info("Initializing...")

        intents = Intents.default()
        intents.members = True
        super().__init__(intents=intents)

        self.storage_dir = storage_dir
        self.plugin_directories = [Path.cwd() / "plugins"]
        if not argv.portable:
            self.plugin_directories.append(self.storage_dir / "plugins")

        self.config_manager = ConfigManager(storage_dir / "config")
        self.config = self.config_manager.config

        self.channel_manager = ChannelManager(self)
        self.command_dispatcher = CommandDispatcher(self)

        self.plugin_manager = PluginManager(self)
        self.plugin_manager.load_all_plugins(self.plugin_directories)

        self.logged_in = False
        self.last_error = None

    async def on_ready(self):
        if not self.logged_in:
            self.logged_in = True
            self.logger.info("Logged in as:")
            self.logger.info(self.user)
            self.logger.info(self.user.id)
            self.logger.info("------------")
            self.logger.info("Activating plugins.")
            create_task(self.global_tick_dispatcher())
            if len(self.guilds) == 0:
                self.logger.info("It looks like you haven't yet added the bot to any servers. Paste the link below "
                                 "into your browser and invite the bot to some servers!")
                self.logger.info(oauth_url(self.user.id))
            await self.plugin_manager.activate_all()

    async def close(self):
        self.logger.warning("Logging out and shutting down.")
        await self.plugin_manager.deactivate_all()
        self.config_manager.save_config()
        await super().close()

    async def on_error(self, event_method, *pargs, **kwargs):
        exc = exc_info()
        self.last_error = exc
        self.logger.exception(f"Unhandled {exc[0].__name__} occurred in {event_method}: ", exc_info=True)

    async def on_resumed(self):
        await self.plugin_manager.hook_event("on_resumed")

    async def on_typing(self, channel, user, when):
        if isinstance(channel, DMChannel):
            return
        if self.channel_manager.channel_in_category(channel.guild, "no_read", channel):
            return
        await self.plugin_manager.hook_event("on_typing", channel, user, when)

    async def on_message(self, msg):
        if msg.guild is not None:
            if self.channel_manager.channel_in_category(msg.guild, "no_read", msg.channel):
                return
            await self.command_dispatcher.command_check(msg)
            await self.plugin_manager.hook_event("on_message", msg)
        else:
            await self.command_dispatcher.command_check(msg)
            await self.plugin_manager.hook_event("on_dm_message", msg)

    async def on_message_delete(self, msg):
        if msg.guild is None:
            return
        if self.channel_manager.channel_in_category(msg.guild, "no_read", msg.channel):
            return
        await self.plugin_manager.hook_event("on_message_delete", msg)

    async def on_message_edit(self, before, after):
        if after.guild is None:
            return
        if self.channel_manager.channel_in_category(after.guild, "no_read", after.channel):
            return
        await self.plugin_manager.hook_event("on_message_edit", before, after)

    async def on_reaction_add(self, reaction, user):
        if reaction.message.guild is None:
            return
        if self.channel_manager.channel_in_category(reaction.message.guild, "no_read", reaction.message.channel):
            return
        await self.plugin_manager.hook_event("on_reaction_add", reaction, user)

    async def on_raw_reaction_add(self, payload):
        if payload.channel_id is not None \
                and self.channel_manager.channel_in_category(
                    Object(payload.guild_id),
                    "no_read",
                    Object(payload.channel_id)):
            return
        await self.plugin_manager.hook_event("on_raw_reaction_add", payload)

    async def on_reaction_remove(self, reaction, user):
        if reaction.message.guild is None:
            return
        if self.channel_manager.channel_in_category(reaction.message.guild, "no_read", reaction.message.channel):
            return
        await self.plugin_manager.hook_event("on_reaction_remove", reaction, user)

    async def on_raw_reaction_remove(self, payload):
        if payload.channel_id is not None \
                and self.channel_manager.channel_in_category(
                    Object(payload.guild_id),
                    "no_read",
                    Object(payload.channel_id)):
            return
        await self.plugin_manager.hook_event("on_raw_reaction_remove", payload)

    async def on_reaction_clear(self, message, reactions):
        if message.guild is None:
            return
        if self.channel_manager.channel_in_category(message.guild, "no_read", message.channel):
            return
        await self.plugin_manager.hook_event("on_reaction_clear", message, reactions)

    async def on_private_channel_create(self, channel):
        await self.plugin_manager.hook_event("on_private_channel_create", channel)

    async def on_private_channel_delete(self, channel):
        await self.plugin_manager.hook_event("on_private_channel_delete", channel)

    async def on_private_channel_update(self, before, after):
        await self.plugin_manager.hook_event("on_private_channel_update", before, after)

    async def on_private_channel_pins_update(self, channel, last_pin):
        await self.plugin_manager.hook_event("on_private_channel_pins_update", channel, last_pin)

    async def on_guild_channel_create(self, channel):
        await self.plugin_manager.hook_event("on_guild_channel_create", channel)

    async def on_guild_channel_delete(self, channel):
        await self.plugin_manager.hook_event("on_guild_channel_delete", channel)

    async def on_guild_channel_update(self, before, after):
        await self.plugin_manager.hook_event("on_guild_channel_update", before, after)

    async def on_guild_channel_pins_update(self, channel, last_pin):
        if self.channel_manager.channel_in_category(channel.guild, "no_read", channel):
            return
        await self.plugin_manager.hook_event("on_guild_channel_pins_update", channel, last_pin)

    async def on_member_join(self, member):
        await self.plugin_manager.hook_event("on_member_join", member)

    async def on_member_remove(self, member):
        await self.plugin_manager.hook_event("on_member_remove", member)

    async def on_member_update(self, before, after):
        await self.plugin_manager.hook_event("on_member_update", before, after)

    async def on_guild_join(self, guild):
        self.channel_manager.add_guild(str(guild.id))
        await self.plugin_manager.hook_event("on_guild_join", guild)

    async def on_guild_remove(self, guild):
        await self.plugin_manager.hook_event("on_guild_remove", guild)

    async def on_guild_update(self, before, after):
        await self.plugin_manager.hook_event("on_guild_update", before, after)

    async def on_guild_role_create(self, role):
        await self.plugin_manager.hook_event("on_guild_role_create", role)

    async def on_guild_role_delete(self, role):
        await self.plugin_manager.hook_event("on_guild_role_delete", role)

    async def on_guild_role_update(self, before, after):
        await self.plugin_manager.hook_event("on_guild_role_update", before, after)

    async def on_guild_emojis_update(self, guild, before, after):
        await self.plugin_manager.hook_event("on_guild_emojis_update", guild, before, after)

    async def on_guild_available(self, guild):
        self.channel_manager.add_guild(str(guild.id))
        await self.plugin_manager.hook_event("on_guild_available", guild)

    async def on_guild_unavailable(self, guild):
        await self.plugin_manager.hook_event("on_guild_unavailable", guild)

    async def on_voice_state_update(self, member, before, after):
        await self.plugin_manager.hook_event("on_voice_state_update", member, before, after)

    async def on_member_ban(self, guild, member):
        await self.plugin_manager.hook_event("on_member_ban", guild, member)

    async def on_member_unban(self, guild, member):
        await self.plugin_manager.hook_event("on_member_unban", guild, member)

    async def on_group_join(self, channel, user):
        await self.plugin_manager.hook_event("on_group_join", channel, user)

    async def on_group_remove(self, channel, user):
        await self.plugin_manager.hook_event("on_group_remove", channel, user)

    async def global_tick_dispatcher(self):
        timer = self.config.get("global_tick_interval", 15)
        while self.logged_in:
            await sleep(timer)
            create_task(self.plugin_manager.hook_event("on_global_tick", datetime.utcnow(), timer))



import asyncio
import json
import re
import shlex
import urllib.request
import urllib.error
from io import BytesIO
from red_star.plugin_manager import BasePlugin
from red_star.rs_errors import CommandSyntaxError, UserPermissionError
from red_star.rs_utils import respond, is_positive, RSArgumentParser, split_message
from red_star.command_dispatcher import Command
from discord import InvalidArgument, HTTPException
from traceback import format_exception, format_exc


class BotManagement(BasePlugin):
    name = "bot_management"
    version = "1.3"
    author = "medeor413"
    description = "A plugin that allows bot maintainers to interface with core bot options through Discord."

    async def on_dm_message(self, msg):
        if msg.author == self.client.user:
            return
        maintainers = [self.client.get_user(x) for x in self.config_manager.config.get("bot_maintainers", [])]
        maintainers = filter(None, maintainers)
        if msg.content.startswith("!") and msg.author in maintainers:
            return
        for user in maintainers:
            await user.send(f"`{msg.author}:` {msg.system_content}\n"
                            f"`Attachments: {', '.join(a.url for a in msg.attachments)}`")

    @Command("Ping",
             doc="Returns the current message latency of the bot.",
             syntax="N/A",
             dm_command=True)
    async def _ping(self, msg):
        await respond(msg, f"**ANALYSIS: Current latency is {round(self.client.latency * 1000)} ms.**")

    @Command("Shutdown",
             doc="Shuts down the bot.",
             category="bot_management",
             dm_command=True,
             bot_maintainers_only=True)
    async def _shutdown(self, msg):
        await respond(msg, "**AFFIRMATIVE. SHUTTING DOWN.**")
        raise SystemExit

    @Command("UpdateAvatar",
             doc="Updates the bot's avatar.",
             syntax="(URL or file)",
             category="bot_management",
             bot_maintainers_only=True,
             dm_command=True)
    async def _update_avatar(self, msg):
        try:
            url = msg.content.split(None, 1)[1]
            img = urllib.request.urlopen(url).read()
        except IndexError:
            if msg.attachments:
                fp = BytesIO()
                await msg.attachments[0].save(fp)
                img = fp.getvalue()
            else:
                raise CommandSyntaxError("No URL or file provided.")
        except (urllib.error.URLError, ValueError):
            raise CommandSyntaxError("Invalid URL provided.")
        try:
            await self.client.user.edit(avatar=img)
            await respond(msg, "**AVATAR UPDATED.**")
        except InvalidArgument:
            raise CommandSyntaxError("Image must be a PNG or JPG")
        except HTTPException:
            raise UserPermissionError("Cannot change avatar at this time.")

    @Command("UpdateName",
             doc="Updates the bot's (nick)name.",
             syntax="[-p/--permanent] (name)",
             category="bot_management",
             bot_maintainers_only=True,
             dm_command=True)
    async def _update_name(self, msg):
        args = msg.clean_content.split()[1:]
        edit_username = False
        if args[0].lower() in ("-p", "--permanent"):
            args.pop(0)
            edit_username = True
        elif args[0].lower() == "reset":
            await msg.guild.me.edit(nick=None)
            await respond(msg, "**ANALYSIS: Nickname reset.**")
            return
        newname = " ".join(args)
        if edit_username:
            if "bot_maintainers" not in self.config_manager.config:
                raise UserPermissionError("No bot maintainers are set!")
            if msg.author.id not in self.config_manager.config["bot_maintainers"]:
                raise UserPermissionError
            await self.client.user.edit(username=newname)
            await respond(msg, f"**ANALYSIS: Username changed to {newname} successfully.**")
        else:
            await msg.guild.me.edit(nick=newname)
            await respond(msg, f"**ANALYSIS: Nickname changed to {newname} successfully.**")

    @Command("Activate",
             doc="Activates an inactive plugin.",
             syntax="(plugin) [permanent]",
             category="bot_management",
             bot_maintainers_only=True,
             dm_command=True)
    async def _activate(self, msg):
        plgname = msg.content.split()[1]
        try:
            permanent = is_positive(msg.content.split()[2])
        except IndexError:
            permanent = False
        all_plugins = self.plugin_manager.plugins
        if plgname in all_plugins:
            if plgname not in self.plugins:
                if plgname in self.config_manager.config["disabled_plugins"] and permanent:
                    self.config_manager.config["disabled_plugins"].remove(plgname)
                    self.config_manager.save_config()
                await self.plugin_manager.activate(plgname)
                await respond(msg, f"**ANALYSIS: Plugin {plgname} was activated successfully.**")
            else:
                await respond(msg, f"**ANALYSIS: Plugin {plgname} is already activated.**")
        else:
            await respond(msg, f"**WARNING: Could not find plugin {plgname}.**")

    @Command("Deactivate",
             doc="Deactivates an active plugin.",
             syntax="(plugin) [permanent]",
             category="bot_management",
             bot_maintainers_only=True,
             dm_command=True)
    async def _deactivate(self, msg):
        plgname = msg.content.split()[1].lower()
        try:
            permanent = is_positive(msg.content.split()[2])
        except IndexError:
            permanent = False
        if plgname == self.name:
            await respond(msg, f"**WARNING: Cannot deactivate {self.name}.**")
        elif plgname in self.plugins:
            if plgname not in self.config_manager.config["disabled_plugins"] and permanent:
                self.config_manager.config["disabled_plugins"].append(plgname)
                self.config_manager.save_config()
            await self.plugin_manager.deactivate(plgname)
            await respond(msg, f"**ANALYSIS: Plugin {plgname} was deactivated successfully.**")
        else:
            await respond(msg, f"**ANALYSIS: Plugin {plgname} is not active.**")

    @Command("ReloadPlugin",
             doc="Reloads a plugin module, refreshing code changes.",
             syntax="(plugin)",
             category="bot_management",
             bot_maintainers_only=True,
             dm_command=True)
    async def _reload_plugin(self, msg):
        plgname = msg.content.split()[1].lower()
        if plgname == self.name:
            await respond(msg, f"**WARNING: Cannot deactivate {self.name}.**")
        elif plgname in self.plugins:
            await self.plugin_manager.reload_plugin(plgname)
            await respond(msg, f"**ANALYSIS: Plugin {plgname} was reloaded successfully.**")

    @Command("ListPlugins",
             doc="Lists all plugins and their activation status.",
             syntax="(plugin)",
             category="bot_management",
             bot_maintainers_only=True,
             dm_command=True)
    async def _list_plugins(self, msg):
        active_plgs = ", ".join(self.plugins.keys())
        if not active_plgs:
            active_plgs = "None."
        all_plgs = list(self.plugin_manager.plugins)
        inactive_plgs = ", ".join([x for x in all_plgs if x not in self.plugins])
        if not inactive_plgs:
            inactive_plgs = "None."
        await respond(msg, f"**ANALYSIS: Plugins are as follows:**```\nActive: {active_plgs}\n"
                           f"Inactive: {inactive_plgs}\n```")

    @Command("GetConfig",
             doc="Gets the config value at the specified path. Use <server> to fill in the server ID.\n"
                 "Use --file to view a different configuration file, such as music_player.json.",
             syntax="(path/to/value) [-f/--file file]",
             category="bot_management",
             bot_maintainers_only=True,
             dm_command=True)
    async def _get_config(self, msg):
        args = msg.clean_content.split()[1:]
        if args[0] in ("-f", "--file"):
            args.pop(0)
            filename = args.pop(0)
            try:
                conf_dict = self.config_manager.plugin_config_files[filename]
            except KeyError:
                raise CommandSyntaxError(f"Config file {filename} does not exist.")
        else:
            conf_dict = self.config_manager.config.copy()

        try:
            path = args[0]
        except IndexError:
            path = ""
        try:
            path = path.replace("<server>", str(msg.guild.id))
        except AttributeError:
            path = path.replace("<server>", "default")
        if path.startswith("/"):
            path = path[1:]
        path_list = path.split("/")
        conf_dict.pop("token", None)  # Don't wanna leak that by accident!

        for k in path_list:
            if not k:
                break
            try:
                conf_dict = self._list_or_dict_subscript(conf_dict, k)
            except KeyError:
                raise CommandSyntaxError(f"Key {k} does not exist.")
            except IndexError:
                raise CommandSyntaxError(f"Index {k} does not exist.")
            except ValueError:
                raise CommandSyntaxError(f"{k} is not a valid integer index.")
            except TypeError:
                raise CommandSyntaxError(f"{args.path} is not a valid path!")

        res = json.dumps(conf_dict, indent=2, sort_keys=True)
        for split_msg in split_message(f"**ANALYSIS: Contents of {path}:**```JSON{res}```"):
            await respond(msg, split_msg)

    @Command("SetConfig",
             doc="Edits the config value at the specified path. Use <server> to fill in the server ID. Doesn't allow "
                 "types to be changed unless forced.\n"
                 "Use --file to edit a different configuration file, such as music_player.json.\n"
                 "Use --remove to delete a value, --append to add a value to a list, --addkey to add a new key/value"
                 "pair to a dict, and --type with a type name to force type conversion.\n"
                 "Valid --types: bool, int, float, str, list, dict, json",
             syntax="(path/to/edit) (value or -r/--remove) [-a/--append] [-k/--addkey key_name] [-t/--type type]"
                    "[-f/--file file]",
             category="bot_management",
             bot_maintainers_only=True,
             dm_command=True)
    async def _set_config(self, msg):
        parser = RSArgumentParser()
        parser.add_argument("path")
        parser.add_argument("value", nargs="*")
        parser.add_argument("-f", "--file", type=str)
        parser.add_argument("-t", "--type", choices=("null", "bool", "int", "float", "str", "list", "dict", "json"),
                            type=str.lower)
        group = parser.add_mutually_exclusive_group()
        group.add_argument("-r", "--remove", action="store_true")
        group.add_argument("-a", "--append", action="store_true")
        group.add_argument("-k", "--addkey")

        args = parser.parse_args(shlex.split(msg.clean_content)[1:])

        type_converters = {
            "null": lambda x: None,
            "bool": is_positive,
            "int": int,
            "float": float,
            "str": str,
            "list": json.loads,
            "dict": json.loads,
            "DotDict": json.loads,
            "json": json.loads
        }

        if args.file:
            try:
                conf_dict = self.config_manager.plugin_config_files[args.file]
            except KeyError:
                raise CommandSyntaxError(f"Config file {args.file} does not exist.")
        else:
            conf_dict = self.config_manager.config.copy()

        try:
            path = args.path
            try:
                path = path.replace("<server>", str(msg.guild.id))
            except AttributeError:
                path = path.replace("<server>", "default")
            if path.startswith("/"):
                path = path[1:]
            path_list = path.split("/")
            final_key = path_list.pop()
            value = " ".join(args.value)
            if args.type:
                try:
                    value = type_converters[args.type](value)
                except ValueError:
                    raise CommandSyntaxError(f"{value} cannot be converted to {args.type}!")
            if value is None and not args.remove and args.type != "null":
                raise CommandSyntaxError("Missing new config value.")
        except IndexError:
            raise CommandSyntaxError("Missing path to config value.")
        for k in path_list:
            try:
                conf_dict = self._list_or_dict_subscript(conf_dict, k)
            except KeyError:
                raise CommandSyntaxError(f"Key {k} does not exist.")
            except IndexError:
                raise CommandSyntaxError(f"Index {k} does not exist.")
            except TypeError:
                raise CommandSyntaxError(f"{args.path} is not a valid path!")

        try:
            orig = self._list_or_dict_subscript(conf_dict, final_key)
        except KeyError:
            raise CommandSyntaxError(f"Key {final_key} does not exist.")
        except IndexError:
            raise CommandSyntaxError(f"Index {final_key} does not exist.")
        except ValueError:
            raise CommandSyntaxError(f"{final_key} is not a valid integer index.")
        except TypeError:
            raise CommandSyntaxError(f"{args.path} is not a valid path!")

        if args.remove:
            if isinstance(conf_dict, dict):
                del conf_dict[final_key]
            elif isinstance(conf_dict, list):
                conf_dict.pop(int(final_key))
            else:
                raise CommandSyntaxError("Path does not lead to a collection!")
            await respond(msg, f"**ANALYSIS: Config value {args.path} deleted successfully.**")

        elif args.append:
            if isinstance(conf_dict[final_key], list):
                conf_dict[final_key].append(value)
                await respond(msg, f"**ANALYSIS: {value} appended to {path} successfully.**")
            else:
                raise CommandSyntaxError("Path does not lead to a list!")

        elif args.addkey:
            if isinstance(conf_dict[final_key], dict):
                conf_dict[final_key][args.addkey] = value
                await respond(msg, f"**ANALYSIS: Key {args.addkey} with value {value} added to {path} "
                                   f"successfully.**")
            else:
                raise CommandSyntaxError("Path does not lead to a dict!")

        elif not args.type and type(value) is not type(orig) and orig is not None:
            orig_type = type(orig).__name__
            try:
                conf_dict[final_key] = type_converters[orig_type](value)
                await respond(msg, f"**ANALYSIS: Config value {path} edited to** `{value}` **successfully.**")
            except (KeyError, ValueError):
                raise CommandSyntaxError(f"Couldn't coerce {value} into type {orig_type}!")
        else:
            conf_dict[final_key] = value
            await respond(msg, f"**ANALYSIS: Config value {path} edited to** `{value}` **successfully.**")

        self.config_manager.save_config()

    @Command("LastError",
             doc="Gets the last error to occur in the specified context.",
             syntax="(command/event/unhandled)",
             category="debug",
             perms={"manage_guild"},
             dm_command=True)
    async def _last_error(self, msg):
        try:
            args = msg.clean_content.split(None, 1)[1].lower()
        except IndexError:
            raise CommandSyntaxError("No error context specified.")
        if args == "command":
            e = self.client.command_dispatcher.last_error
        elif args == "event":
            e = self.plugin_manager.last_error
        elif args == "unhandled":
            e = self.client.last_error
        else:
            raise CommandSyntaxError("Invalid error context.")
        if e:
            excstr = "\n".join(format_exception(*e))
            await respond(msg, f"**ANALYSIS: Last error in context {args}:** ```Python\n{excstr}\n```")
        else:
            await respond(msg, f"**ANALYSIS: No error in context {args}.**")

    @Command("Execute", "Exec", "Eval",
             doc="Executes the given Python code. Be careful, you can really break things with this!\n"
                 "Provided variables are `ct` (shorthand for asyncio.create_task) and `self.res` "
                 "(printed by the bot after execution).",
             syntax="(code in code block)",
             category="debug",
             bot_maintainers_only=True,
             run_anywhere=True,
             dm_command=True)
    async def _execute(self, msg):
        try:
            arg = re.split(r"\s+", msg.content, 1)[1]
            t_match = re.match(r"`([^\n\r`]+)`|```.*?\s+(.*)```", arg, re.DOTALL)
            code = t_match.group(1) or t_match.group(2)
        except IndexError:
            raise CommandSyntaxError("No code provided.")
        except AttributeError:
            raise CommandSyntaxError("Code is not in valid code block.")
        # Convenience variables for printing results and create_task
        ct = asyncio.create_task
        self.res = self.EmptySentinel()
        # noinspection PyBroadException
        try:
            exec(code, globals(), locals())
        except Exception:
            await respond(msg, f"**WARNING: Error occurred while executing code. Traceback:**\n"
                               f"```Py\n{format_exc()}\n```")
        if not isinstance(self.res, self.EmptySentinel):
            await respond(msg, f"**ANALYSIS: Result: {self.res}**")
            self.res = self.EmptySentinel()

    @staticmethod
    def _list_or_dict_subscript(obj, key):
        """
        Helper function to use key as a key if obj is a dict, or as an int index if list.

        :param obj: The object to subscript.
        :param key: The key to subscript with.
        :return: The subscripted object.
        """
        if isinstance(obj, dict):
            return obj[key]
        elif isinstance(obj, list):
            return obj[int(key)]
        else:
            raise TypeError

    class EmptySentinel:
        """
        Just a simple class to function as a sentinel value for Execute's result variable.
        """



# Error classes in a neutral space to avoid import order issues.


class CommandSyntaxError(Exception):
    # We don't want to catch code-based syntax errors by accident
    pass


class UserPermissionError(Exception):
    # For when a command user lacks the permissions needed.
    # We don't want to catch OS permission errors by accident, unlikely as they are
    pass


class ChannelNotFoundError(TypeError):
    # For when a channel of type x doesn't exist
    pass


class CustomCommandSyntaxError(CommandSyntaxError):
    # For when the CC author made a syntax error
    pass


class ConsoleCommandSyntaxError(CommandSyntaxError):
    # For errors in console commands
    pass


class DataCarrier(Exception):
    # This is intended to carry a message up out of a stack, not to signal any actual error.
    def __init__(self, data):
        self.data = data



import json
import logging
import sys
from pathlib import Path
from shutil import copyfile
from red_star.rs_utils import JsonFileDict


class ConfigManager:
    """
    Manages the loading and modification of the configuration files.
    """
    def __init__(self, config_path: Path):
        self.logger = logging.getLogger("red_star.config_manager")
        self.logger.debug("Initialized config manager.")
        self.config = {}
        self.config_path = config_path
        self.config_file_path = config_path / "config.json"
        self.plugin_config_files = {}
        self.load_config()

    def load_config(self):
        temp_path = Path(str(self.config_file_path) + "_bak")
        self.logger.debug("Loading configuration...")
        try:
            with self.config_file_path.open(encoding="utf-8") as fd:
                self.config = json.load(fd)
        except FileNotFoundError:
            if temp_path.exists():
                temp_path.rename(self.config_file_path)
                self.logger.warning(f"Couldn't find {self.config_file_path}!\n"
                                    f"A backup from an interrupted save was found. Attempting to load...")
                return self.load_config()
            self.logger.warning(f"Couldn't find {self.config_file_path}! Copying default configuration...")
            default_path = Path.cwd() / "_default_files/config.json.default"
            self.config_path.mkdir(parents=True, exist_ok=True)
            copyfile(str(default_path), str(self.config_file_path))
            self.logger.info(f"A default configuration has been copied to {self.config_path}.\n"
                             f"Please configure the bot before continuing.")
            sys.exit(1)
        except json.decoder.JSONDecodeError:
            if temp_path.exists():
                self.config_file_path.unlink()
                temp_path.rename(self.config_file_path)
                self.logger.warning(f"The configuration file located at {self.config_file_path} is invalid!\n"
                                    f"A backup from an interrupted save was found. Attempting to load...")
                return self.load_config()
            self.logger.exception(f"The configuration file located at {self.config_file_path} is invalid!\n"
                                  f"Please correct the error below and restart.", exc_info=True)
            sys.exit(1)

        if "plugins" not in self.config:
            self.config["plugins"] = {}

    def save_config(self):
        temp_path = Path(str(self.config_file_path) + "_bak")
        with temp_path.open("w", encoding="utf-8") as f:
            json.dump(self.config, f, sort_keys=True, indent=2)
        self.config_file_path.unlink()
        temp_path.rename(self.config_file_path)
        for file in self.plugin_config_files.values():
            file.save()
        self.logger.debug("Saved config files.")

    def get_plugin_config(self, name):
        if name not in self.config["plugins"]:
            self.config["plugins"][name] = {}
        conf = self.config["plugins"][name]
        return conf

    def init_plugin_config(self, name, default_config):
        new_config = default_config.copy()
        current_config = self.config["plugins"].get(name, {})
        new_config.update(current_config)
        self.config["plugins"][name] = new_config

    def get_plugin_config_file(self, filename, json_save_args=None, json_load_args=None) -> JsonFileDict:
        if filename in self.plugin_config_files:
            file_obj = self.plugin_config_files[filename]
        else:
            file_path = self.config_path / filename
            if not file_path.exists():
                default_config = Path.cwd() / "_default_files" / (filename + ".default")
                if default_config.exists():
                    copyfile(str(default_config), str(file_path))
                    self.logger.debug(f"Copied default configuration for {filename} to {file_path}.")
                else:
                    with file_path.open("w", encoding="utf-8") as fd:
                        fd.write("{}")
                    self.logger.debug(f"Created config file {file_path}.")
            file_obj = JsonFileDict(file_path, json_save_args, json_load_args)
            self.plugin_config_files[filename] = file_obj
        return file_obj

    def is_maintainer(self, member):
        return member.id in self.config.get('bot_maintainers', [])
